/**
 *
 * Copyright (c) 2000 Justin E. Forrester (forrestje@bigfoot.com)
 * All Rights Reserved
 * 
 * I provide the Windows "Fuzz" Testing Tool on an AS IS basis, and
 * do not warrant its validity or performance.  I reserve the right
 * to update, modify, or discontinue this software at any time.  I
 * shall have no obligation to supply such updates or modifications or
 * any other form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. I define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Fuzz or parts of Fuzz, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without my permission. For
 * other uses, please contact me at forrestje@bigfoot.com. 
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Fuzz, you understand and agree that I (or any
 * other person or entity with proprietary rights in Fuzz) am
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Fuzz.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall I (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * me (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Fuzz.
 *
 **/

#include "Fuzz.h"
#include "WindowList.h"


//=============================================================================
// MAIN
//=============================================================================
int main(int argc, char *argv[])
{
    LONG nErrStatus = NO_ERROR;

    // create and zero out a struct for the fuzz testing options
    FUZZ_OPTIONS options;
    ZeroMemory(&options, sizeof(FUZZ_OPTIONS));

    //================================================
    // DEAL WITH COMMAND LINE ARGS
    //================================================

    // parse the command line arguments
    nErrStatus = parseArgs(argc, argv, &options);
    if (NO_ERROR != nErrStatus) 
    { 
        fprintf(stderr, "Invalid command line\n");
        printHelp();
        return nErrStatus; 
    }

    // validate the necessary command line args
    if (!options.pszAppCmdLine && 0 == options.dwProcessId)
    {
        fprintf(stderr, "Invalid command line or pid\n");
        printHelp();
        return ERR_INVALID_CMD_LINE;
    }
    if (! options.fRandomWin32ViaSendMessage && 
        ! options.fRandomWin32ViaPostMessage &&
        ! options.fRandomEvents)
    {
        fprintf(stderr, "No test specified\n");
        printHelp();
        return ERR_INVALID_CMD_LINE;
    }

    // hide the console window away for the duration of message sending
    // to ensure that the test environment is as sterile as possible
    // i.e. there is no possibility of interaction with windows other than
    // the system windows which are common to all systems (e.g. desktop, etc)
    HWND hConsoleWindow = GetForegroundWindow();
    WINDOWPLACEMENT wp;
    ZeroMemory(&wp, sizeof(WINDOWPLACEMENT));
    wp.length = sizeof(WINDOWPLACEMENT);
    GetWindowPlacement(hConsoleWindow, &wp);
    wp.showCmd = SW_MINIMIZE;
    SetWindowPlacement(hConsoleWindow, &wp);

    //================================================
    // START TARGET APPLICATION
    //================================================

    // fire up the target application
    HANDLE hTargetProcess = INVALID_HANDLE_VALUE;
    if (0 == options.dwProcessId) // pid not spec'd on command line
    {
        // start up the application to be tested
        PROCESS_INFORMATION piTargetInfo;
        nErrStatus = launchApp(options.pszAppCmdLine, piTargetInfo);
        if (NO_ERROR != nErrStatus) 
        { 
            fprintf(stderr, "Error launching app\n");

            // done with the tests, restore the console window
            wp.showCmd = SW_RESTORE;
            SetWindowPlacement(hConsoleWindow, &wp);

            return nErrStatus; 
        }
        hTargetProcess = piTargetInfo.hProcess;

        // update the options with the correct process id
        options.dwProcessId = piTargetInfo.dwProcessId;

        // sleep for a while and let the testee finish initilazing
        fprintf(stdout, "Pausing to allow target to load...");
        Sleep(LAUNCH_WAIT_TIME);
        //    WaitForInputIdle(piTargetInfo.hProcess, LAUNCH_WAIT_TIME);
    }
    else // pid spec'd on command line
    {
        // get process information
        hTargetProcess = OpenProcess(PROCESS_QUERY_INFORMATION, 
                                     FALSE, 
                                     options.dwProcessId);
        if (NULL == hTargetProcess)
        {
            fprintf(stderr, "Unable to get process information for pid %d\n", options.dwProcessId);

            // done with the tests, restore the console window
            wp.showCmd = SW_RESTORE;
            SetWindowPlacement(hConsoleWindow, &wp);

            return ERR_INVALID_PID;
        }
    }
    
    //================================================
    // GET WINDOW INFO FOR TARGET APPLICATION
    //================================================

    // get the top level window for the target application
    WindowList *pWindowList = new WindowList();
    pWindowList->genWindowList(options.dwProcessId);

    // make sure that we found at least 1 window
    LONG nWindowCount = pWindowList->numWindows();
    if (0 == nWindowCount)
    {
        fprintf(stderr, "Unable to get top level window for process id %d\n", options.dwProcessId);
        
        // done with the tests, restore the console window
        wp.showCmd = SW_RESTORE;
        SetWindowPlacement(hConsoleWindow, &wp);    

        return ERR_WND_NOT_FOUND;
    }

    HWND hTargetWindow = (HWND) INVALID_HANDLE_VALUE;
    LONG nTargetIndex = 1;

    // if there are more than 1 top-level windows associated with 
    // this process, present the tester with a list to choose from, 
    // since we will be sending messages to only 1 top level window
    if (nWindowCount > 1)
    {
        // pop the console window back up
        wp.showCmd = SW_RESTORE;
        SetWindowPlacement(hConsoleWindow, &wp);

        // present list of entries to choose from
        fprintf(stdout, "\n\nMore than one top-level window found for process %d\n", options.dwProcessId);
        fprintf(stdout, "Choose target for tests:\n\n");

        fprintf(stdout, "%2s  %-25.25s  %-20.20s  (%-10.10s) (%-10.10s)\n", 
                "#",
                "Window caption",
                "Window class",
                "Upper Left",
                "Lower Rt.");

        fprintf(stdout, "%2s  %-25.25s  %-20.20s  %-12s %-12s\n", 
                "--",
                "-------------------------",
                "--------------------",
                "------------",
                "------------");

        // print the list of windows
        for (int i = 1; i <= nWindowCount; i++)
        {
            WindowSpec *pCurr = pWindowList->getWindowAt(i);
            
            fprintf(stdout, "%2d. %-25.25s  %-20.20s  (%-4d, %-4d) (%-4d, %-4d)\n", 
                    i, 
                    pCurr->pszCaption,
                    pCurr->pszWndClassName,
                    pCurr->rect.left,
                    pCurr->rect.top,
                    pCurr->rect.right,
                    pCurr->rect.bottom
               );
        }

        // read the user input
        do 
        {
            fprintf(stdout, "\nEnter # of choice: ");
            fflush(stdin);
        } while ((0 == fscanf(stdin, "%d", &nTargetIndex)) ||
                 (nTargetIndex < 0) || 
                 (nTargetIndex > nWindowCount));
        
        // hide the console window again
        wp.showCmd = SW_MINIMIZE; 
        SetWindowPlacement(hConsoleWindow, &wp);
    }

    // set the target window
    hTargetWindow = pWindowList->getWindowAt(nTargetIndex)->hWnd;

    //================================================
    // DO MESSAGE SENDING
    //================================================

    // seed the random number generator
    if (0 == options.dwSeed)
    {
        srand((unsigned)time(NULL));
    }
    else
    {
        srand(options.dwSeed);
    }

    // do message sending
    fprintf(stdout, "Sending to pid %d, hwnd %#x\n", 
            options.dwProcessId, 
            hTargetWindow);
    if (options.fRandomEvents)
    {
        sendRandomEvents(hTargetWindow, hTargetProcess, &options);

        // clean up key and mouse state
        cleanupMouseKeyboardState();
    }
    else // do pure random tests
    {
        sendRandomWin32(hTargetWindow, hTargetProcess, &options);
    }

    // done with the tests, restore the console window
    wp.showCmd = SW_RESTORE;
    SetWindowPlacement(hConsoleWindow, &wp);

    fprintf(stdout, "Test complete\n");

    return NO_ERROR;

} // end main


//=============================================================================
// Function: parseArgs
//     Description:
//         naive parse routine to handle command line arguments
//     Parameters:
//         nNumArgs - number of arguments
//         ppszArgs - array of string arguments
//         options - structure to fill in with command line options
//     Return Value:
//         handle to the window 
//         on failure, INVALID_HANDLE_VALUE will be returned
//=============================================================================
LONG parseArgs(LONG nNumArgs, LPSTR ppszArgs[], PFUZZ_OPTIONS pOptions)
{
    // validate params
    _ASSERT(ppszArgs);
    _ASSERT(pOptions);

    // Note: this is a VERY dumb parser, but anything more complex is not 
    // needed in a simple command line app just for arguments.

    // make sure the user entered enough arguments
    if (1 == nNumArgs)
    {
        printHelp();
        exit(1);
    }

    // parse arguments
    for (int i = 1; i < nNumArgs; i++)
    {
        LPSTR pszCurrArg = ppszArgs[i]; 
        
        if (0 == strcmp(pszCurrArg, "-?") ||
            0 == strcmp(pszCurrArg, "/?"))
        {
            printHelp();
            exit(1);
        }
        else if (0 == strcmp(pszCurrArg, "-a"))
        {
            // make sure there is another arg
            if (i+1 >= nNumArgs)
            {
                fprintf(stderr, "-a switch given without app cmd line\n");
                return ERR_INVALID_ARG;
            }

            pOptions->pszAppCmdLine = ppszArgs[i+1];
            
            // skip by the number
            i++;
        }
        else if (0 == strcmp(pszCurrArg, "-ws"))
        {
            pOptions->fRandomWin32ViaSendMessage = TRUE;
        }
        else if (0 == strcmp(pszCurrArg, "-wp"))
        {
            pOptions->fRandomWin32ViaPostMessage = TRUE;
        }
        else if (0 == strcmp(pszCurrArg, "-v"))
        {
            pOptions->fRandomEvents = TRUE;
        }
        else if (0 == strcmp(pszCurrArg, "-n"))
        {
            // make sure there is another arg
            if (i+1 >= nNumArgs)
            {
                fprintf(stderr, "-n switch given without number\n");
                return ERR_INVALID_ARG;
            }

            pOptions->dwNumMessages = (DWORD) atoi(ppszArgs[i+1]);
            
            // skip by the number
            i++;
        }
        else if (0 == strcmp(pszCurrArg, "-c"))
        {
            pOptions->fSendWMCommandMessages = TRUE;
        }
        else if (0 == strcmp(pszCurrArg, "-l"))
        {
            pOptions->fSendNULLParams = TRUE;
        }
        else if (0 == strcmp(pszCurrArg, "-e"))
        {
            // make sure there is another arg
            if (i+1 >= nNumArgs)
            {
                fprintf(stderr, "-e switch given without seed\n");
                return ERR_INVALID_ARG;
            }

            pOptions->dwSeed = (DWORD) atoi(ppszArgs[i+1]);
            
            // skip by the number
            i++;
        }            
        else if (0 == strcmp(pszCurrArg, "-i"))
        {
            // make sure there is another arg
            if (i+1 >= nNumArgs)
            {
                fprintf(stderr, "-i switch given without pid\n");
                return ERR_INVALID_ARG;
            }

            pOptions->dwProcessId = (DWORD) atoi(ppszArgs[i+1]);
            
            // skip by the number
            i++;
        }            
        else
        {
            fprintf(stderr, "Invalid argument: %s\n", pszCurrArg);
        } 
    }
    return NO_ERROR;

} // end parseArgs


//=============================================================================
// Function: launchApp
//     Description:
//         launch process with a given command line
//     Parameters:
//         pszCommandLine - command line to use when launching
//         pi - reference to proc info struct that will receive 
//              information about the newly created process
//     Return Value:
//         0 if process was successfully started
//         non-zero otherwise
//=============================================================================
LONG launchApp(LPSTR pszCommandLine, PROCESS_INFORMATION &pi)
{
    // validate params
    _ASSERT(pszCommandLine);

    STARTUPINFO si;    
    ZeroMemory( &si, sizeof(si) );
    si.cb = sizeof(si);    

    // Start the child process. 
    if( !CreateProcess( NULL, // No module name (use command line). 
                        pszCommandLine,   // Command line. 
                        NULL,  // Process handle not inheritable. 
                        NULL,  // Thread handle not inheritable. 
                        FALSE, // Set handle inheritance to FALSE. 
                        0,     // No creation flags. 
                        NULL,  // Use parent's environment block. 
                        NULL,  // Use parent's starting directory. 
                        &si,   // Pointer to STARTUPINFO structure.
                        &pi))  // Ptr to PROCESS_INFORMATION structure. 
    {
        // error creating process, notify user
        LPVOID lpMsgBuf;
        FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                      FORMAT_MESSAGE_FROM_SYSTEM,    
                      NULL,
                      GetLastError(),
                      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
                      (LPTSTR) &lpMsgBuf,
                      0,
                      NULL );
        fprintf(stderr, "Error launching new application: %s\n", pszCommandLine);
        LocalFree( lpMsgBuf );
        return ERR_BAD_LAUNCH;
    }
    return NO_ERROR;

} // end launchApp


//=============================================================================
// Function: sendRandomWin32
//     Description:
//         Perform random win32 message test
//     Parameters:
//         hwnd - window to send messages to
//         hProc - handle to the target process
//         pOptions - testing options
//     Return Value:
//         error status
//=============================================================================
LONG sendRandomWin32(HWND hwnd, HANDLE hProc, PFUZZ_OPTIONS pOptions)
{
    // validate params
    _ASSERT(pOptions);

    if (hwnd == INVALID_HANDLE_VALUE)
    {
        return ERR_INVALID_ARG;
    }
    
    // message decls
    UINT Msg = 0;
    UINT wParam = 0;
    UINT lParam = 0;
    
    DWORD dwNumSent = 0;
    DWORD dwNumTimeouts = 0;
    while (dwNumSent < pOptions->dwNumMessages || 0 == pOptions->dwNumMessages)
    {
        // make sure process is still alive
        DWORD dwExitStatus = 0;
        if (0 == GetExitCodeProcess(hProc, &dwExitStatus))
        {
            return ERR_INVALID_ARG;
        }
        else if (STILL_ACTIVE != dwExitStatus)
        {
            fprintf(stdout, "The target was closed or killed\n");
            break;
        }

        // generate random message type
        do 
        {
            if (pOptions->fSendWMCommandMessages)
                Msg = WM_COMMAND;
            else
                Msg = (UINT) rand();
        
            if (pOptions->fSendNULLParams)
            {
                wParam = NULL;
                lParam = NULL;
            }
            else
            {
                wParam = (UINT) rand();
                lParam = (LONG) rand();
            }
        }
        // don't send close, quit, or destroy msgs
        while (WM_CLOSE == Msg || WM_QUIT == Msg || WM_DESTROY == Msg);
        
        if (pOptions->fRandomWin32ViaPostMessage)
        {
            PostMessage(hwnd, Msg, wParam, lParam);
            dwNumSent++;

        }
        else if (pOptions->fRandomWin32ViaSendMessage)
        {
            // we are also using SendMessageTimeout to avoid any deadlock
            // conditions where the app is waiting for input that will
            // never come because we are waiting for the app to process the 
            // message -- ex. right-click context menus that are popped
            // up by the WM_CONTEXTMENU message
            
            // An Access Violation occasionally occurred in User32.dll 
            // after SendMessage returned.  Documentation in MSDN 
            // indicates this is not just a problem with this app.  
            // The error was very intermittent and not critical, so the 
            // exception is ignored. 
            __try 
            { 
                DWORD dwResult;
                if (FAILED(SendMessageTimeout(hwnd, Msg, wParam, lParam, 
                                              0, 5000, &dwResult)))
                    dwNumTimeouts++;
            } 
            __except(EXCEPTION_EXECUTE_HANDLER) { }

            dwNumSent++;
            
        }

        // send out a status message every 50,000 messages
        if (dwNumSent % 50000 == 0)
        {
            fprintf(stdout, "Number of messages sent: %d\n", dwNumSent);
        }

        // let us play nice with other processes and yield
        Sleep(0);
    }
    if (dwNumTimeouts > 0)
        fprintf(stdout, "Number of timeouts: %d\n", dwNumTimeouts);
    fprintf(stdout, "Number of messages sent: %d\n", dwNumSent);
    return NO_ERROR;

} // end sendRandomWin32


//=============================================================================
// Function: sendRandomEvents
//     Description:
//         Perform random event test
//     Parameters:
//         hwnd - window to send messages to
//         hProc - handle to the target process
//         pOptions - testing options
//     Return Value:
//         error status
//=============================================================================
LONG sendRandomEvents(HWND hwnd, HANDLE hProc, PFUZZ_OPTIONS pOptions)
{
    // allocate a local RECT for window dimensions
    RECT rectWindowDimensions;

    // allocate array to keep track of key states
    // index 0 is unused as there are no VK codes of 0
    BOOL pfKeyStates[255];
    ZeroMemory(pfKeyStates, 255*sizeof(BOOL));

    // allocate array to keep track of mouse states
    BOOL pfMouseButtonStates[2];
    ZeroMemory(pfKeyStates, 2*sizeof(BOOL));

    // validate params
    if (!IsWindow(hwnd) ||
        INVALID_HANDLE_VALUE == hProc ||
        !pOptions)
    {
        return ERR_INVALID_ARG;
    }
        
    // determine current screen resolution for normalizing
    // mouse event coordinates
    DEVMODE modeInfo;
    ZeroMemory(&modeInfo, sizeof(DEVMODE));
    modeInfo.dmSize = sizeof(DEVMODE);
    EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &modeInfo);
    DWORD dwScreenHeight = modeInfo.dmPelsHeight;
    DWORD dwScreenWidth = modeInfo.dmPelsWidth;
    
    // choose random event to send
    DWORD dwNumSent = 0;
    while (dwNumSent < pOptions->dwNumMessages || 0 == pOptions->dwNumMessages)
    {
        // make sure process is still alive
        DWORD dwExitStatus = 0;
        if (0 == GetExitCodeProcess(hProc, &dwExitStatus))
        {
            return ERR_INVALID_ARG;
        }
        else if (STILL_ACTIVE != dwExitStatus)
        {
            fprintf(stdout, "The target was closed or killed\n");
            break;
        }

		// make sure that the target window is owned by the target application
		DWORD dwForegroundPId = 0;
		GetWindowThreadProcessId(GetForegroundWindow(), &dwForegroundPId);
        if (dwForegroundPId != pOptions->dwProcessId)
        {
			SetForegroundWindow(hwnd);
        }

        // synthesize random event into system event queue
        switch (rand() % 2)
        {
        case 0: // KEYBOARD EVENT
            generateRandomKeyboardEvent(pfKeyStates);
            break;
        case 1: // MOUSE EVENT
            // if the window has been minimized, bring it back
            // it is useless to test an app w/random keyboard and mouse 
            // input if it is minimized
            WINDOWPLACEMENT wp;
            ZeroMemory(&wp, sizeof(WINDOWPLACEMENT));
            wp.length = sizeof(WINDOWPLACEMENT);
            GetWindowPlacement(hwnd, &wp);
            if (SW_SHOWMINIMIZED == wp.showCmd)
            {
                fprintf(stderr, "window has been minimized, setting normal\n");
                wp.showCmd = SW_RESTORE;
                SetWindowPlacement(hwnd, &wp);
            }
            
            // get dimensions of top level window
            ZeroMemory(&rectWindowDimensions, sizeof(RECT));

            if (0 == GetClientRect(hwnd, &rectWindowDimensions))
            {
                fprintf(stderr, "Could not determine size of window\n");
                return ERR_SIZE_FAIL;
            }
			
			// convert the client area coords to screen coordinates
			ClientToScreen(hwnd, (LPPOINT)&rectWindowDimensions.left);
			ClientToScreen(hwnd, (LPPOINT)&rectWindowDimensions.right); 

            // validate size, i.e. not 0 pixels wide or tall
            if (0 == rectWindowDimensions.right-rectWindowDimensions.left ||
                0 == rectWindowDimensions.bottom-rectWindowDimensions.top)
            {
                fprintf(stderr, "Error! Top level window has dimension 0x0\n");
                return ERR_TOP_LEVEL_ZERO_DIM;
            }

            // fire away!
            generateRandomMouseEvent(pfMouseButtonStates, 
                                     &rectWindowDimensions,
                                     dwScreenHeight, 
                                     dwScreenWidth);
            break;
        }    
        dwNumSent++;

        // give the system some time to deal with the new event
        Sleep(EVENT_WAIT_TIME);
        
        // since this test has a tendency to get out of hand, 
        // give the user a break (ana a chance to kill the test) 
        // every x messages
        if (dwNumSent % 5000 == 0)
        {
			fprintf(stdout, "%d events sent\n", dwNumSent);
        }
        
        
    }

    return NO_ERROR;

} // end sendRandomEvents


//=============================================================================
// Function: generateRandomKeyboardEvent
//     Description:
//         synthesize a random keyboard event into the system event stream
//     Parameters:
//         pfKeyStates - array of key states
//     Return Value:
//         error status
//=============================================================================
LONG generateRandomKeyboardEvent(BOOL *pfKeyStates)
{
    // validate params
    _ASSERT(pfKeyStates);

    // key event params
    BYTE bVKCode = 0;
    BYTE bScanCode = 0;
    DWORD dwFlags = 0;
    DWORD dwExtraInfo = 0;
        
    do // generate random keyboard event information
    {
        // generate random vkey information
        bVKCode = (rand() % 254) + 1;
    }
    while (isSystemHotKey(bVKCode, dwFlags, pfKeyStates));
        
    // fill in rest of even params

    // not using scan code for tests
    
    // if the key is currently down, send an up
    if (pfKeyStates[bVKCode])
    {
        dwFlags = KEYEVENTF_KEYUP;
    }
    pfKeyStates[bVKCode] = !pfKeyStates[bVKCode];

    // not using extra info for tests

    // synthesize event into system event queue
    keybd_event(bVKCode, bScanCode, dwFlags, dwExtraInfo);

    return NO_ERROR;

} // end generateRandomKeyboardEvent
    

//=============================================================================
// Function: generateRandomMouseEvent
//     Description:
//         synthesize a random mouse event into the system event stream
//         simulates a simple two button mouse with no wheel
//     Parameters:
//         pfMouseButtonState - array of mouse button states
//         pWndDim - dimensions of the target window
//     Return Value:
//         error status
//=============================================================================
LONG generateRandomMouseEvent(BOOL *pfMouseButtonStates, RECT *pWndDims,
                              DWORD dwScreenHeight, DWORD dwScreenWidth)
{
    DWORD dwFlags, dwExtraInfo, dwX, dwY, dwWheelData;
    LONG nRetVal = 0;

    // generate coordinates for mouse event
    dwX = (rand() % (pWndDims->right - pWndDims->left)) + pWndDims->left;
    dwY = (rand() % (pWndDims->bottom - pWndDims->top)) + pWndDims->top;

    // normalize the coordinates for call to mouse_event
    // mouse_event expects coordinates modulo 65536
    dwX = (DWORD) ((FLOAT) dwX / dwScreenWidth * 65535);
    dwY = (DWORD) ((FLOAT) dwY / dwScreenHeight * 65535);

    // set flags
    dwFlags = 0;
    switch(rand() % 3)
    {
    case 0:  // LEFT BUTTON EVENT
        if (pfMouseButtonStates[LBUTTON]) 
            dwFlags = MOUSEEVENTF_LEFTDOWN;
        else
            dwFlags = MOUSEEVENTF_LEFTUP;
        pfMouseButtonStates[LBUTTON] = !pfMouseButtonStates[LBUTTON];
        break;
    case 1:  // RIGHT BUTTON EVENT
        if (pfMouseButtonStates[RBUTTON]) 
            dwFlags = MOUSEEVENTF_RIGHTDOWN;
        else
            dwFlags = MOUSEEVENTF_RIGHTUP;
        pfMouseButtonStates[RBUTTON] = !pfMouseButtonStates[RBUTTON];
        break;
    case 2:  // MOUSE MOTION EVENT
        dwFlags = MOUSEEVENTF_MOVE;
        break;
    }
            
    // we are testing absolute mouse coords, set flag indicating that
    dwFlags |= MOUSEEVENTF_ABSOLUTE;

    // not testing wheel actions
    dwWheelData = 0;
    dwExtraInfo = 0;

    // add event to system queue
    mouse_event(dwFlags, dwX, dwY, dwWheelData, dwExtraInfo);

    return NO_ERROR;

} // end generateRandomMouseEvent


//=============================================================================
// Function: cleanupMouseKeyboardState
//     Description:
//         do our best to tidy up after ourselves by making sure that 
//         all the keys and mouse buttons are up
//     Parameters:
//         none
//     Return Value:
//         none
//=============================================================================
VOID cleanupMouseKeyboardState(VOID)
{
    // down & up all the keys
    for (BYTE bVKCode = 0; bVKCode < 255; bVKCode++)
    {
        keybd_event(bVKCode, 0, 0, 0);
        keybd_event(bVKCode, 0, KEYEVENTF_KEYUP, 0);
    }

    // up the mouse buttons
    mouse_event(MOUSEEVENTF_LEFTUP, 0, 0, 0, 0);
    mouse_event(MOUSEEVENTF_RIGHTUP, 0, 0, 0, 0);

} // end cleanupMouseEvent
    

//=============================================================================
// Function: printHelp
//     Description:
//         print help message about program usage
//     Parameters:
//         none
//     Return Value:
//         none
//=============================================================================
VOID printHelp(VOID)
{
    fprintf(stderr, "\n\n");
    fprintf(stderr, "Usage: Fuzz [-ws] [-wp] [-v] [-a application command line] [-i pid] [-n NumMessages] [-c] [-l] [-e seed]\n\n");
    fprintf(stderr, "Requirements:\n");
    fprintf(stderr, "\tOne of -ws, -wp, -v\n");
    fprintf(stderr, "\tOne of -a, -i\n\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "\t-ws\tRandom Win32 message test using SendMessage API function\n");
    fprintf(stderr, "\t-wp\tRandom Win32 message test using PostMessage API function\n");
    fprintf(stderr, "\t-v\tRandom mouse/keyboard event test\n");
    fprintf(stderr, "\t-a\tLaunch and test given application\n");
    fprintf(stderr, "\t-i\tTest running application with given pid\n");
    fprintf(stderr, "\t-n\tNumber of messages to send (infinite if not specified)\n");
    fprintf(stderr, "\t-c\tSend random WM_COMMAND messages (for pure random test only)\n");
    fprintf(stderr, "\t-l\tSend null parameters in messages (for pure random test only)\n");
    fprintf(stderr, "\t-e\tSet random seed to value\n");
    
    fprintf(stderr, "\n\n");

} // end printHelp


//=============================================================================
// Function: isSystemHotKey
//     Description:
//         determine whether the given key is a system hotkey
//         For the random tests, we don't want to introduce known hotkeys that 
//         will things such as the following:
//             - switch control to another application (ALT-TAB)
//             - close the given window (ALT-F4)
//     Parameters:
//         bVKCode - the virtual key code of the key in question
//         dwFlags - key event flags (up or down essentially, not ext)
//         pfKeyStates - context in which the key event occurs
//                       this is needed to tell if context keys are pressed
//     Return Value:
//         TRUE if the given key event will trigger a system hotkey
//         FALSE otherwise
//=============================================================================
BOOL isSystemHotKey(BYTE bVKCode, DWORD dwFlags, BOOL *pfKeyStates)
{
    // validate params
    _ASSERT(pfKeyStates);

    BOOL fIsHotKey = FALSE;

    switch (bVKCode)
    {
    case VK_TAB: // ALT-TAB
        if (pfKeyStates[VK_MENU]) fIsHotKey = TRUE;
        break;

    case VK_F4: // ALT-F4 & CTRL-F4
        if (pfKeyStates[VK_MENU] || pfKeyStates[VK_CONTROL]) fIsHotKey = TRUE;
        break;

	case VK_F1:   // help
    case VK_LWIN: // anything involving the windows keys 
    case VK_RWIN:
        fIsHotKey = TRUE;
        break;

    case VK_ESCAPE: // CTRL-ESC (start menu) and ALT-ESC (task switch)
        if (pfKeyStates[VK_CONTROL] || pfKeyStates[VK_MENU]) fIsHotKey = TRUE;
        break;

    case VK_CONTROL: // CTRL-ALT
        if (pfKeyStates[VK_MENU]) fIsHotKey = TRUE;
        break;

    case VK_MENU: // ALT-CTRL
        if (pfKeyStates[VK_CONTROL]) fIsHotKey = TRUE;
        break;
    }

    return fIsHotKey;

} // end isSystemHotKey

